from __future__ import annotations
from dataclasses import dataclass
import shutil
import subprocess
import fire
import yaml

from uvmono.github import find_projects, get_includes, set_github_output
from uvmono.template import (
    add_project_standards,
    create_devcontainer,
    create_dockerfile,
    sync_pyproject,
)
from uvmono.utils import find_git_root, has_uv


@dataclass
class Include:
    path: str
    name: str
    dependencies: list[str]


class UvMono:
    """uvmono is a tool to help manage mono-repos in python"""

    def __init__(self):
        self._root = find_git_root()
        self._packages_root = self._root / "packages"
        if (
            not self._packages_root.exists()
            and input(
                "Packages directory not found. Would you like to create it? (y/n)"
            ).lower()
            == "y"
        ):
            self._packages_root.mkdir()

        self._packages = [p for p in self._packages_root.iterdir() if p.is_dir()]

    def new(self, package_name: str, overwrite: bool = False):
        """Create a new package in the mono-repo.

        Args:
            package_name (str): The name of the package to create.
            overwrite (bool, optional): Whether to overwrite the package if it already exists.
        """
        has_uv()
        package_dir = self._packages_root / package_name
        if package_dir.exists() and not overwrite and not overwrite:
            raise FileExistsError(f"Package {package_name} already exists")
        elif package_dir.exists() and overwrite:
            print(f"Overwriting package: {package_name}")
            shutil.rmtree(package_dir)
        package_dir.mkdir()
        dir_cmd = ["--directory", str(package_dir)]
        default_packages = ["pytest", "pytest-cov", "ruff", "pre-commit", "commitizen"]
        subprocess.run(["uv", "init", "--package", *dir_cmd])
        assert (package_dir / "pyproject.toml").exists(), (
            f"Failed to create package: {package_name}"
        )
        add_project_standards(package_dir)
        subprocess.run(["uv", "add", *default_packages, "--dev", *dir_cmd])
        self.add_devcontainer(package_name=package_name)

    def sync(self, package_name: str = "", all: bool = False, dry_run: bool = False):
        if not package_name and not all:
            return "Either `package_name` or `all` must be set. Use `--help` for more information."
        packages = self.list() if all else [package_name]
        for package_name in packages:
            package_path = self._packages_root / package_name
            print(f"Syncing pyproject.toml for {package_name}")
            if not package_path.exists():
                raise FileNotFoundError(f"Package {package_name} not found")
            if dry_run:
                print(f"[DRY_RUN] Would sync pyproject.toml for {package_name}")
                continue
            sync_pyproject(package_path, package_name)

    def list(self):
        """List the packages in the mono-repo"""
        packages = [p.name for p in self._packages_root.iterdir() if p.is_dir()]
        return sorted(packages)

    def add_devcontainer(
        self,
        package_name: str = "",
        all: bool = False,
        dry_run: bool = False,
        ubuntu_version: str = "22.04",
        spark_version: str = "3.5.4",
        uv_version: str = "0.5.15",
    ):
        """Add a devcontainer configuration to a package

        Args:
            package_name (str): The name of the package to add the devcontainer to.
            all (bool, optional): Whether to add the devcontainer to all packages.
            dry_run (bool, optional): Whether to run in dry-run mode.
            ubuntu_version (str, optional): The version of ubuntu to use.
            spark_version (str, optional): The version of spark to use.
            uv_version (str, optional): The version of uv to use.
        """
        if not package_name and not all:
            return "Either `package_name` or `all` must be set. Use `--help` for more information."
        packages = self.list() if all else [package_name]
        self._create_dockerfiles(
            ubuntu_version=ubuntu_version,
            spark_version=spark_version,
            uv_version=uv_version,
            dry_run=dry_run,
        )
        self._create_devcontainers(packages, dry_run=dry_run)
        return f"Successfully created devcontainer configuration{'' if all is True else 's'}"

    def _create_devcontainers(self, packages: list[str], dry_run=False):
        print(f"Adding devcontainer configuration to {', '.join(packages)}")
        for package in packages:
            container_path = self._root / ".devcontainer" / package
            dot_to_root = "/".join(
                [".." for _ in container_path.relative_to(self._root).parts]
            )
            if container_path.exists():
                print(
                    f"Devcontainer configuration already exists for {package}. Overwriting"
                )
            docker_compose_file = f"{dot_to_root}/.docker/docker-compose.yml"
            template = create_devcontainer(
                package_name=package,
                dockerfile_path=docker_compose_file,
            )
            if dry_run:
                print(
                    f"[DRY_RUN] Would create devcontainer configuration for {package}"
                )
                continue
            container_path.mkdir(exist_ok=True, parents=True)
            with open(container_path / "devcontainer.json", "w") as f:
                f.write(template)
            print(f"Successfully created devcontainer configuration for {package}")

    def _create_dockerfiles(
        self, ubuntu_version, spark_version, uv_version, dry_run=False
    ):
        compose = {
            "version": "3",
            "services": {
                package: {
                    "build": {
                        "context": "..",
                        "dockerfile": ".docker/Dockerfile",
                        "args": {
                            "UBUNTU_VERSION": ubuntu_version,
                            "SPARK_VERSION": spark_version,
                            "UV_VERSION": uv_version,
                        },
                    },
                    "volumes": ["..:/workspace:cached"],
                    "command": "sleep infinity",
                }
                for package in self.list()
            },
        }
        if dry_run:
            print("Would create docker-compose.yml with the following content:")
            print(yaml.dump(compose))
            return
        if not (self._root / ".docker").exists():
            (self._root / ".docker").mkdir()
        with open(self._root / ".docker" / "docker-compose.yml", "w") as f:
            yaml.dump(compose, f, indent=4)
        dockerfile_path = self._root / ".docker" / "Dockerfile"
        dockerfile_template = create_dockerfile(
            ubuntu_version=ubuntu_version,
            spark_version=spark_version,
            uv_version=uv_version,
        )
        with open(dockerfile_path, "w") as f:
            f.write(dockerfile_template)

    def matrix_strategy(self, key: str, dry_run: bool = False) -> list[Include]:
        """Set the matrix strategy for the mono-repo

        Args:
            key (str): The key to use for the matrix strategy.
                This also appends the key to the `GITHUB_OUTPUT` environment variable.
            dry_run (bool, optional): Whether to run in dry-run mode. Defaults to False.

        Example:
            $ uvmono matrix-strategy python
        """
        projects = find_projects()
        includes = get_includes(projects)
        if not dry_run:
            set_github_output(key, includes)
        return includes

    def _get_relative_package_path(self, package_name: str):
        """Get the path to the package"""
        package_path = self._packages_root / package_name
        if not package_path.exists():
            raise FileNotFoundError(f"Package {package_name} not found")
        return package_path.relative_to(self._packages_root.parent).as_posix()


def main() -> None:
    fire.Fire(UvMono)


if __name__ == "__main__":
    main()



from pathlib import Path
import tomllib
from jinja2 import Environment, FileSystemLoader
import mergedeep
import tomli_w

TEMPLATES_DIR = Path(__file__).parent / "templates"


def sync_pyproject(package_dir: Path, package_name: str):
    tomlpath = package_dir / "pyproject.toml"
    template_file = _render("pyproject.toml.j2", package_name=package_name)
    updates = tomllib.loads(template_file)
    with open(tomlpath, "rb") as f:
        current = tomllib.load(f)
    updated = mergedeep.merge(current, updates, strategy=mergedeep.Strategy.REPLACE)
    with open(tomlpath, "wb") as f:
        tomli_w.dump(updated, f)


def add_project_standards(package_dir: Path):
    package_name = package_dir.name.replace("-", "_")
    _create_tests(package_dir / "tests", package_name)
    _add_pyproject_defaults(package_dir, package_name)


def create_filter(project_path: Path, dependents: list[Path]) -> str:
    deps = [(p.path / "**").as_posix() for p in dependents]
    if not deps:
        return _render("filter.yml.j2", path=project_path.as_posix())
    return _render("filter.yml.j2", path=project_path.as_posix(), dependencies=deps)


def _render(template_name: str, **kwargs) -> str:
    """Render a jinja2 template

    Args:
        template_name (str): The name of the template to render
        **kwargs: Keyword arguments to pass to the template

    Returns:
        str: The rendered template
    """
    env = Environment(loader=FileSystemLoader(TEMPLATES_DIR))
    template = env.get_template(template_name)
    return template.render(**kwargs)


def create_dockerfile(
    ubuntu_version: str,
    spark_version: str,
    uv_version: str,
):
    return _render(
        "Dockerfile.j2",
        ubuntu_version=ubuntu_version,
        spark_version=spark_version,
        uv_version=uv_version,
    )


def create_devcontainer(
    package_name: str,
    dockerfile_path: Path,
):
    return _render(
        "devcontainer.json.j2",
        package_name=package_name,
        docker_compose_file=dockerfile_path,
    )


def _create_tests(tests_dir: Path, package_name: str):
    main_template = _render("tests_init.py.j2", package_name=package_name)
    conftest_template = _render("conftest.py.j2", package_name=package_name)
    tests_dir.mkdir()
    (tests_dir / "__init__.py").write_text("")
    (tests_dir / "test_main.py").write_text(f"{main_template}\n")
    (tests_dir / "conftest.py").write_text(f"{conftest_template}\n")


def _add_pyproject_defaults(package_dir: Path, package_name: str):
    init_template = _render("package_init.py.j2", package_name=package_name)
    init_template_path = package_dir / "src" / package_name / "__init__.py"
    init_template_path.write_text(f"{init_template}\n")
    with open(package_dir / "pyproject.toml", "a") as fh:
        pyproject_template = _render("pyproject.toml.j2", package_name=package_name)
        fh.write(pyproject_template)



from uvmono.github import get_dependents, Project, Dependency


def test_get_dependents():
    p1 = Project(
        name="project1",
        version="1.0.0",
        path="path1",
        dependencies=[],
    )
    p2 = Project(
        name="project2",
        version="1.0.0",
        path="path2",
        dependencies=[Dependency("project1", is_workspace=True)],
    )
    p3 = Project(
        name="project3",
        version="1.0.0",
        path="path3",
        dependencies=[Dependency("project2", is_workspace=True)],
    )
    projects = [p1, p2, p3]

    dependents = get_dependents(projects)
    assert dependents == {
        "project1": [],
        "project2": [p1],
        "project3": [p2, p1],
    }



from pathlib import Path
import shutil
import subprocess


def find_git_root(root=__file__) -> Path:
    """Find the root of the git repository

    Returns:
        root_path (Path): The root of the git repository
    """
    root_path = Path(root).parent
    max_depth = 50
    while root_path.is_dir() and not (root_path / ".git").exists() and max_depth > 0:
        root_path = root_path.parent
        max_depth -= 1
    if not (root_path / ".git").exists():
        raise RuntimeError("Could not find git root")
    return root_path


def has_uv():
    """Check if `uv` is installed"""
    has_uv = shutil.which("uv") or subprocess.run(["uv", "--version"]).returncode == 0
    if not has_uv:
        msg = """`uv` not found.

        This package requires the `uv` command line tool.

        You can find information on how to install it here:
            https://docs.astral.sh/uv/getting-started/installation/ 
        """
        raise RuntimeError(msg)
    return True



from dataclasses import dataclass
import json
import os
from pathlib import Path
import tomllib

from uvmono.template import create_filter


@dataclass
class Dependency:
    name: str
    is_workspace: bool


@dataclass
class Project:
    name: str
    version: str
    path: Path
    dependencies: list[Dependency]


def find_projects(root="."):
    projects = {}
    for toml in Path(root).rglob("pyproject.toml"):
        with open(toml, "rb") as fh:
            data = tomllib.load(fh)

        project = data.get("project")
        if not project:
            raise ValueError(f"Missing project in {toml}")

        name = project.get("name")
        version = project.get("version")
        if not name or not version:
            raise ValueError(
                f"Missing name or version in {toml} | name={name}, version={version}"
            )

        project_dependencies = project.get("dependencies", [])
        uv_sources = data.get("tool", {}).get("uv", {}).get("sources", {})
        dependencies = []
        for dep in project_dependencies:
            dependencies.append(Dependency(dep, dep in uv_sources))
        projects[name] = Project(name, version, toml.parent, dependencies)

    return projects


def get_includes(projects: dict[str, Project]):
    includes = []
    dependents = get_dependents(list(projects.values()))

    for project in projects.values():
        include = {
            "path": project.path.as_posix(),
            "name": project.name,
            "dependencies": json.dumps(
                [(p.path / "**").as_posix() for p in dependents.get(project.name, [])]
            ),
            "filter": create_filter(project.path, dependents.get(project.name, [])),
        }
        print(f"------- {project.name}\n{include['filter']}")

        includes.append(include)

    return includes


def set_github_output(key: str, value: list[dict[str, str]]):
    include_statement = json.dumps({key: value})
    msg = f"matrix={include_statement}"
    print(f"Set GITHUB_OUTPUT to:\n{msg}")
    try:
        with open(os.environ["GITHUB_OUTPUT"], "a") as fh:
            print(msg, file=fh)
    except KeyError:
        raise RuntimeError("GITHUB_OUTPUT environment variable not set")


def _get_dependents(project: Project, projects: list[Project]):
    dependents = []
    for dep in project.dependencies:
        dep_project = next((p for p in projects if p.name == dep.name), None)
        if dep_project:
            dependents.append(dep_project)
            dependents.extend(_get_dependents(dep_project, projects))
    return dependents


def get_dependents(projects: list[Project]):
    dependents: dict[str, list[Project]] = {project.name: [] for project in projects}
    for project in projects:
        dependents[project.name] = _get_dependents(project, projects)
    return dependents
